# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import logging
import libgiza.config

logger = logging.getLogger('giza.config.deploy')


class DeployConfig(libgiza.config.ConfigurationBase):
    @property
    def production(self):
        return self.state['production']

    @production.setter
    def production(self, value):
        self.state['production'] = DeployTargetConfig(value)

    @property
    def testing(self):
        return self.state['testing']

    @testing.setter
    def testing(self, value):
        self.state['testing'] = DeployTargetConfig(value)

    @property
    def docsprod(self):
        return self.state['docsprod']

    @docsprod.setter
    def docsprod(self, value):
        self.state['docsprod'] = DeployTargetConfig(value)

    def get_staging(self, project):
        """Returns the appropriate StagingTargetConfig staging configuration for
           the current project."""
        try:
            return self.state['staging'][project]
        except KeyError as err:
            logger.critical('No staging information specified for project %s',
                            project)
            raise err

    def get_deploy(self, project):
        """Returns the appropriate StagingTargetConfig deployment configuration
           for the current project."""
        try:
            return self.state['s3_deploy'][project]
        except KeyError as err:
            logger.critical('No deployment information specified for project %s',
                            project)
            raise err

    @property
    def staging(self):
        return self.state['staging']

    @staging.setter
    def staging(self, value):
        self.state['staging'] = {}
        for project, config in value.items():
            self.state['staging'][project] = StagingTargetConfig(config)

    @property
    def s3_deploy(self):
        return self.state['s3_deploy']

    @s3_deploy.setter
    def s3_deploy(self, value):
        self.state['s3_deploy'] = {}
        for project, config in value.items():
            self.state['s3_deploy'][project] = StagingTargetConfig(config)


class DeployTargetConfig(libgiza.config.ConfigurationBase):
    _option_registry = ['input']

    @property
    def args(self):
        return self.state['args']

    @args.setter
    def args(self, value):
        if isinstance(value, list):
            self.state['args'] = value
        else:
            logger.critical('deployment arguments must be a list')
            raise TypeError

    @property
    def hosts(self):
        return self.state['hosts']

    @hosts.setter
    def hosts(self, value):
        if isinstance(value, list):
            self.state['hosts'] = value
        else:
            logger.critical('deployment targets must be a list')
            raise TypeError


class StagingTargetConfig(libgiza.config.ConfigurationBase):
    """Configuration for a project's staging environment, specifying both an S3
       bucket and an HTTP/S URL pointing to that bucket, appropriate for user
       consumption."""
    @property
    def url(self):
        try:
            return self.state['url']
        except KeyError as err:
            logger.critical('No staging URL specified')
            raise err

    @url.setter
    def url(self, value):
        self.state['url'] = str(value)

    @property
    def bucket(self):
        try:
            return self.state['bucket']
        except KeyError as err:
            logger.critical('No staging bucket specified')
            raise err

    @bucket.setter
    def bucket(self, value):
        self.state['bucket'] = str(value)

    @property
    def prefix(self):
        """The prefix underwhich giza will upload everything in the bucket."""
        try:
            return self.state['prefix']
        except KeyError as err:
            return ''

    @prefix.setter
    def prefix(self, value):
        self.state['prefix'] = str(value)

    @property
    def use_branch(self):
        """Specifies whether only the directory corresponding to the current
           git branch should be uploaded. False if giza should upload
           everything."""
        try:
            return self.state['use_branch']
        except KeyError as err:
            return True

    @use_branch.setter
    def use_branch(self, value):
        self.state['use_branch'] = bool(value)

    @property
    def redirect_dirs(self):
        """A list of directory regular expressions for which this project
           is considered to "own" redirects. Redirects outside of these
           directories will never be removed (but may be modified) by
           this project."""
        try:
            return self.state['redirect_dirs']
        except KeyError as err:
            return []

    @redirect_dirs.setter
    def redirect_dirs(self, value):
        self.state['redirect_dirs'] = [re.compile(d) for d in value]



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from giza.inheritance import InheritableContentBase

logger = logging.getLogger('giza.content.tocs.models')

if sys.version_info >= (3, 0):
    basestring = str


class TocData(InheritableContentBase):
    _option_registry = ['edition', 'name', 'description']

    @property
    def file(self):
        return self.state['file']

    @file.setter
    def file(self, value):
        if not isinstance(value, basestring):
            logger.error("filenames must be strings: " + str(value))
        elif not value.startswith("/"):
            logger.error("'{0}' is not a valid file specification".format(value))
            raise TypeError
        elif value.endswith(".rst") or value.endswith(".txt"):
            logger.error("file specifications cannot end with extensions: " + value)
            raise TypeError
        else:
            self.state['file'] = value

    @property
    def ref(self):
        if 'ref' not in self.state:
            self.ref = self.file

        return self.state['ref']

    @ref.setter
    def ref(self, value):
        self.state['ref'] = value

    @property
    def level(self):
        if 'level' not in self.state:
            return 1
        else:
            return self.state['level']

    @level.setter
    def level(self, value):
        if isinstance(value, int):
            self.state['level'] = value
        else:
            self.state['level'] = int(value)

    @property
    def text_only(self):
        if 'text_only' not in self.state:
            return False
        else:
            return self.state['text_only']

    @text_only.setter
    def text_only(self, value):
        if isinstance(value, bool):
            self.state['text_only'] = value
        else:
            self.state['text_only'] = bool(value)

    @property
    def is_spec(self):
        if self.source is None:
            return False
        else:
            return True



# 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper functions and data for interacting with content generation and
processing. The most used function is
:fun:`~giza.content.helper.edition_check()` which compares a content object and
the current build environment and returns a boolean that reflects if the object
is appropriate for the edition.
"""

from pygments.lexers import get_all_lexers

level_characters = {
    "=": 1,
    "-": 2,
    "~": 3,
    "`": 4,
    "^": 5,
    "'": 6
}

character_levels = dict(zip(level_characters.values(),
                            level_characters.keys()))


def edition_check(data, conf):
    """
    Tests a content structure against the current configuration object to ensure
    that the editions match. Used so that content generation scripts can
    consistently filter out content from non-matching editions.

    :param dict,object data: A dictionary or object that holds content.

    :param giza.config.main.Configuration conf: A Configuration object.

    :returns: ``True`` if ``data`` does not contain an ``edition`` field or
       attribute of ``data`` matches the value of
       ``conf.project.edition``. ``False`` if the edition does not match
       ``conf.project.edition`` or if ``conf.project.edition`` is the same as
       the project name (i.e. ``conf.project.name``).

    :rtype: Boolean
    """

    if 'edition' in data:
        # compatibility to support data that are objects as well as (typically
        # legacy) dictionary data.
        try:
            local_edition = data.edition
        except AttributeError:
            local_edition = data['edition']

        # if the edition is the project's short name, assume that there is
        # effectively no edition, and consider that all editions match.
        if conf.project.edition == conf.project.name:
            return True

        if isinstance(local_edition, list):
            # if the content item has more than one edition, the current edition
            # setting only needs to be present in that list.

            return conf.project.edition in local_edition
        else:
            # if the content item specifies a single (string) edition name, it
            # must match the current edition string.

            return local_edition == conf.project.edition
    else:
        # if the content does not specify an edition, then the build will render
        # all data.

        return True

# get a list of all supported pygment lexers.


def get_all_languages():
    all_languages = ['none']

    for lexers in get_all_lexers():
        all_languages.extend(lexers[1])

    return all_languages



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import shutil

from libgiza.task import Task

from giza.content.glossary.inheritance import GlossaryDataCache
from giza.content.glossary.views import render_glossary
from giza.config.content import new_content_type

logger = logging.getLogger('giza.content.extract.tasks')


def register_glossary(conf):
    content_dfn = new_content_type(name='glossary',
                                   task_generator=glossary_tasks,
                                   conf=conf)

    conf.system.content.add(name='glossary', definition=content_dfn)


def write_glossary(terms, fn):
    content = render_glossary(terms)
    content.write(fn)
    logger.info('wrote glossary file: ' + fn)


def glossary_tasks(conf):
    terms = GlossaryDataCache(conf.system.content.glossary.sources, conf)
    terms.create_output_dir()

    tasks = []
    for fn, glossary_file in terms.file_iter():
        tasks.append(Task(job=write_glossary,
                          args=(glossary_file, glossary_file.target(fn)),
                          description="generate glossary for: " + fn,
                          target=glossary_file.target(fn),
                          dependency=fn))

    logger.info("add {0} glossary tasks".format(len(tasks)))
    return tasks


def glossary_clean(conf):
    return [Task(job=shutil.rmtree,
                 args=[conf.system.content.glossary.output_dir],
                 target=True,
                 dependency=[conf.system.content.glossary.output_dir],
                 description='removing {0}'.format(conf.system.content.glossary.output_dir))]



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module creates captures hashes of all content in the repository and stores
this information between builds, so that the build system has the ability to
test which files change in between builds, so that ``giza`` does not need to
rely on ``mtime``, which can change following branch operations in git.

Furthermore the tasks created as part of the build here use the implicit
relationship between source files determined by :mod:`giza.includes` in
combination with this cache of file hashes to ensure that Sphinx does not skip
building files.
"""

import datetime
import json
import logging
import os

import libgiza.task

from giza.includes import include_files
from giza.tools.files import expand_tree, safe_create_directory, md5_file
from giza.tools.timing import Timer

logger = logging.getLogger('giza.content.dependencies')

# Update File Hashes


def dump_file_hashes(conf):
    output = conf.system.dependency_cache

    o = {'time': datetime.datetime.utcnow().strftime("%s"),
         'files': {}}

    files = expand_tree(os.path.join(conf.paths.projectroot, conf.paths.branch_source), None)

    fmap = o['files']

    for fn in files:
        if os.path.exists(fn):
            fmap[fn] = md5_file(fn)

    safe_create_directory(os.path.dirname(output))

    with open(output, 'w') as f:
        json.dump(o, f)

    logger.info('wrote dependency cache to: {0}'.format(output))

# Update Dependencies


def _refresh_deps(graph, dep_map, conf):
    warned = set()
    count = 0

    # For each file in the source tree, bump the timestamp of all files that
    # include it, if the file changed since the last build.

    for file, dependents in graph.items():
        if check_hashed_dependency(file, dep_map, conf) is True:
            core_file = normalize_dep_path(file, conf, False)
            norm_file = normalize_dep_path(file, conf, True)

            if os.path.isfile(norm_file) and not os.path.isfile(core_file):
                # these are generated files in the build/<branch>/source. No
                # need to touch these files.
                continue
            for dep in [normalize_dep_path(dep, conf, branch=True) for dep in dependents]:
                if not os.path.exists(core_file):
                    # this file doesn't exist in the source. Sphinx will
                    # warn about this file later (though the output silently
                    # ignores the unavailable content.)

                    if core_file in warned:
                        continue
                    else:
                        warned.add(core_file)
                        logger.warning('included file does not exist: ' + core_file)
                elif os.path.exists(dep):
                    logger.debug('updating timestamp of "{0}" because of "{1}"'.format(dep, file))
                    os.utime(dep, None)
                    count += 1

    logger.info('bumped timestamps for {0} files'.format(count))


def refresh_deps(conf):
    with Timer('resolve dependency graph'):
        # resolve a map of the source files to the files they depend on
        # (i.e. the ones that they include).
        graph = include_files(conf=conf)

        # load, if possible, a mappping of all source files with hashes from the
        # last build.
        if not os.path.exists(conf.system.dependency_cache):
            dep_map = None
        else:
            with open(conf.system.dependency_cache, 'r') as f:
                try:
                    dep_cache = json.load(f)
                    dep_map = dep_cache['files']
                except ValueError:
                    dep_map = None
                    m = 'no stored dependency information, will rebuild more things than necessary.'
                    logger.warning(m)

    with Timer('dependency updates'):
        _refresh_deps(graph, dep_map, conf)

# In previous versions, giza loaded the dep_map in the main thread, and then
# passed the checks and update to a worker pool, but the pool took ~40 seconds
# on a large resource, and doing the entire operation serially in a thread takes
# ~1 second.


def refresh_dependency_tasks(conf):
    return [libgiza.task.Task(job=refresh_deps,
                              args=[conf],
                              target=None,
                              dependency=conf.system.dependency_cache,
                              description="check and touch files affected by dependency changes")]


def dump_file_hash_tasks(conf):
    return [libgiza.task.Task(job=dump_file_hashes,
                              args=[conf],
                              target=conf.system.dependency_cache,
                              dependency=os.path.join(conf.paths.projectroot,
                                                      conf.paths.branch_source),
                              description="writing dependency cache to a file for the next build")]

# Hashed Dependency Checking


def normalize_dep_path(fn, conf, branch):
    """
    Given a filename (typically, absolute with regards to the ``source``
    directory), the configuration object, the ``branch`` boolean, returns a
    fully qualified path of a source file.

    When ``branch`` is ``False``, that's in the actual ``source/``
    directory. When ``branch`` is ``True``, that's the proxy-source directory in
    ``build/<branch>``.
    """

    fn = fn.rstrip()
    if not fn.startswith(conf.paths.projectroot):
        if fn.startswith(conf.paths.branch_source) or fn.startswith(conf.paths.source):
            fn = os.path.join(conf.paths.projectroot, fn)
        elif fn.startswith('/'):
            if branch is False:
                fn = os.path.join(conf.paths.projectroot,
                                  conf.paths.source,
                                  fn[1:])
            else:
                fn = os.path.join(conf.paths.projectroot,
                                  conf.paths.branch_source,
                                  fn[1:])

    return fn


def check_hashed_dependency(fn, dep_map, conf):
    """
    :return: ``True`` when any of the files that include ``fn`` have changed since
        the generation of the the ``dep_map``. Always returns ``True`` if
        ``dep_map`` is ``None`` (i.e. if this is the first build.)
    """
    # logger.info('checking dependency for: ' + fan)

    fn = normalize_dep_path(fn, conf, branch=False)

    if dep_map is None:
        return True
    elif not os.path.exists(fn):
        return True
    elif fn in dep_map:
        if dep_map[fn] != md5_file(fn):
            return True
        else:
            return False
    else:
        return False



# 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from libgiza.config import RecursiveConfigurationBase

logger = logging.getLogger('giza.config.sphinx_config')


class ReplacementData(RecursiveConfigurationBase):

    def ingest(self, input_obj):
        if isinstance(input_obj, list):
            if len(input_obj) == 1 and isinstance(input_obj[0], dict):
                input_obj = input_obj[0]
            else:
                try:
                    input_obj = dict((item['edition'], item) for item in input_obj)
                except KeyError:
                    logger.error("replacement specification is malformed. documents need editions")
                    return

        if self.conf.project.edition == self.conf.project.name:
            if self.conf.project.name in input_obj:
                self._update_tokens(input_obj[self.conf.project.name])
            else:
                if self._validate_tokens(input_obj) is True:
                    self._update_tokens(input_obj)

        if self.conf.project.edition in input_obj:
            self._update_tokens(input_obj[self.conf.project.edition])
        else:
            logger.error('current edition not defined for replacements, adding no replacements')

    def _validate_tokens(self, tokens):
        for value in tokens.items():
            if isinstance(value, dict):
                logger.error("replacement tokens cannot specify mappings")
                return False

        return True

    def _update_tokens(self, new_keys):
        if 'tokens' in new_keys:
            self.state.update(new_keys['tokens'])
        else:
            self.state.update(new_keys)

    def items(self):
        return self.state.items()

    def keys(self):
        return self.state.keys()

    def values(self):
        return self.state.values()

    def update(self, value):
        self._update_tokens(value)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

from giza.inheritance import InheritableContentBase
from giza.content.steps.models import HeadingMixin, ActionMixin
from giza.content.helper import get_all_languages

logger = logging.getLogger('giza.content.models')


class ReleaseData(HeadingMixin, ActionMixin, InheritableContentBase):
    _option_registry = ['copyable', 'ref', 'description', 'pre', 'post', 'content']

    @property
    def target(self):
        return os.path.join(self.conf.system.content.releases.fn_prefix, self.ref) + '.rst'



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Alternate implementation of the system that fetches intersphinx inventories from
other Sphinx sites. Implemented separately to avoid using Sphinx so that that
Giza can check and update inventories once per build-run rather than once per
build run. Useful for minimizing the start-up time for ``sphinx-build`` and
useful for reducing redundant work in parallel build situations.
"""

import time
import os
import logging
import subprocess

import libgiza.task

from giza.tools.files import verbose_remove, safe_create_directory

logger = logging.getLogger('giza.content.intersphinx')

ACCEPTABLE = 864000

# Helper functions


def download_file(file, url):
    cmd = ['curl', '--silent', '--location', '--remote-time', url, '-o', file]

    safe_create_directory(os.path.dirname(file))

    try:
        subprocess.check_call(cmd)
        logger.info('downloaded {0}'.format(file))
        return True
    except subprocess.CalledProcessError:
        logger.error('trouble downloading interspinx inventory: ' + file)
        return False


def file_timestamp(path):
    return os.stat(path)[8]

# Tasks


def download(f, s, conf):
    if conf.runstate.force is True:
        newf = download_file(f, s)

    if os.path.isfile(f):
        newf = False
    else:
        logger.info('{0} file does not exist, downloading now'.format(f))
        newf = download_file(f, s)

        if newf is False:
            m = "intersphinx inventory ({0}) download failed. skipping"
            logger.warning(m.format(f))
            return

    mtime = file_timestamp(f)

    if mtime < time.time() - ACCEPTABLE:
        # if mtime is less than now - n days, it may be stale.

        newtime = time.time() - (ACCEPTABLE / 2)

        if newf is True:
            # if we just downloaded the file it isn't stale yet
            os.utime(f, (newtime, newtime))
        else:
            # definitley stale, must download it again.
            newf = download_file(f, s)
            if mtime == file_timestamp(f):
                # if the source is stale, modify mtime so we don't
                # download it for a few days.
                os.utime(f, (newtime, newtime))


def intersphinx_tasks(conf):
    if 'intersphinx' not in conf.system.files.data:
        return

    tasks = []
    for i in conf.system.files.data.intersphinx:
        try:
            f = os.path.join(conf.paths.projectroot,
                             conf.paths.output, i.path)

            s = i.url + 'objects.inv'
        except AttributeError:
            f = os.path.join(conf.paths.projectroot,
                             conf.paths.output, i['path'])

            s = i['url'] + 'objects.inv'

        description = 'download intersphinx inventory from {0}'.format(s)
        tasks.append(libgiza.task.Task(job=download,
                                       args=(f, s, conf),
                                       target=f,
                                       dependency=None,
                                       description=description))
        logger.debug('added job for {0}'.format(s))

    return tasks


def intersphinx_clean(conf):
    tasks = []

    for inv in conf.system.files.data.intersphinx:
        try:
            fn = os.path.join(conf.paths.projectroot,
                              conf.paths.output, inv.path)
        except AttributeError:
            fn = os.path.join(conf.paths.projectroot,
                              conf.paths.output, inv['path'])

        if os.path.exists(fn):
            t = libgiza.task.Task(job=verbose_remove,
                                  args=[fn])
            tasks.append(t)

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Diagnostic operation that displays the majority of the current configuration
object.
"""

import json
import argh

import giza
import libgiza

from giza.config.helper import fetch_config


@argh.named('version')
@argh.expects_obj
def report_version(args):
    "Returns current version of giza"

    print("giza: " + giza.__version__)
    print("libgiza: " + libgiza.__version__)


@argh.arg('--conf_path', '-c')
@argh.arg('--edition', '-e')
@argh.arg('--language', '-l')
@argh.named('config')
@argh.expects_obj
def render_config(args):
    """Returns configuration object for diagnostic purposes."""

    c = fetch_config(args)

    # the following values are rendered lazily. we list them here so that the
    # final object will be useful to inspect.
    [c.git.commit, c.paths.public, c.git.branches.current,
     c.git.branches.manual, c.git.branches.published,
     c.paths.branch_output, c.paths.buildarchive,
     c.paths.branch_source, c.paths.branch_staging,
     c.paths.branch_images, c.paths.branch_includes,
     c.version.published, c.version.stable, c.version.upcoming,
     c.project.edition, c.deploy, c.paths.global_config,
     c.project.branched, c.system.dependency_cache,
     c.system.dependency_cache_fn, c.paths.public_site_output,
     c.system.content, c.runstate.runner, c.runstate.force,
     c.system.files, c.system.files.paths, c.system.files.data,
     c.paths.htaccess]

    print('--- ' + "str of config object >>>")
    print(json.dumps(c.dict(), indent=3))
    print('---  <<<')



import giza

from setuptools import setup, find_packages

REQUIRES = [
    'argh',
    'docutils',
    'hieroglyph',
    'jinja2',
    'libgiza==0.2.13',
    'onetimepass',
    'polib',
    'pyyaml',
    'rstcloth==0.2.6',
    'sphinx-intl',
    'sphinx==1.2.3',
    'sphinxcontrib-httpdomain',
    'wand',
    'boto',
    'fett',
    'jira'
]

setup(
    name='giza',
    maintainer='tychoish',
    maintainer_email='sam@tychoish.com',
    description='Sphinx Documentation Build Automation',
    version=giza.__version__,
    license='Apache 2.0',
    url='http://github.com/mongodb/docs-tools.git',
    packages=find_packages(),
    test_suite=None,
    install_requires=REQUIRES,
    package_data={'giza': ['quickstart/makefile', "quickstart/source/*", "quickstart/source/.gitignore", 'quickstart/config/*.yaml']},
    extras_require={
        'jira': ['jira', 'pyOpenSSL', 'ndg-httpsclient', 'pyasn1', 'requests>=2.1.0'],
        'github': ['github3.py']
    },
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Documentation',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'License :: OSI Approved :: Apache Software License',
    ],
    entry_points={
        'console_scripts': [
            'giza = giza.cmdline:main',
            'scrumpy = giza.scrumpy:main [jira]',
            'mdbpr = giza.github:main [github]'
           ],
        },
    )



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
import re
import libgiza.config
import giza.content.helper

logger = logging.getLogger('giza.config.migrations')

if sys.version_info >= (3, 0):
    basestring = str


class MigrationData(list):
    def append(self, item):
        self.insert(-1, item)

    def extend(self, items):
        for item in items:
            self.insert(-1, item)

    def insert(self, index, item):
        if item is None:
            return
        elif giza.content.helper.edition_check(item, self.conf) is False:
            return

        if 'sources' in item:
            spec = {}
            if 'source_dir' in item:
                spec['source_dir'] = item['source_dir']
            for source in item['sources']:
                spec['source'] = source
                m = MigrationSpecification(spec, self.conf)
                super(MigrationData, self).insert(-1, m)
        else:
            m = MigrationSpecification(item, self.conf)
            super(MigrationData, self).insert(index, m)


class MigrationSpecification(libgiza.config.RecursiveConfigurationBase):
    @property
    def source_dir(self):
        if 'source_dir' not in self.state:
            self.source_dir = None

        return self.state['source_dir']

    @source_dir.setter
    def source_dir(self, value):
        paths = []
        if value is not None:
            if value.startswith('/'):
                value = value[1:]
            paths.extend([
                value,
                os.path.join(self.conf.paths.projectroot, value),
                os.path.join(self.conf.paths.projectroot, self.conf.paths.output, value)])

        paths.extend([self.conf.paths.projectroot,
                      os.path.abspath(os.path.join(self.conf.paths.projectroot, '..')),
                      os.path.abspath(os.path.join(self.conf.paths.projectroot, '..', 'source')),
                      os.path.join(self.conf.paths.projectroot, self.conf.paths.source),
                      os.path.join(self.conf.paths.projectroot, self.conf.paths.output)])
        self.state['source_dir'] = paths

    @property
    def source(self):
        return self.state['source']

    @source.setter
    def source(self, value):
        if value.startswith('/'):
            value = value[1:]

        self.state['_raw_source'] = value

        fns = [os.path.join(path, value)
               for path in self.source_dir]

        for fn in fns:
            if os.path.isfile(fn):
                self.state['source'] = fn
                break

        if 'source' not in self.state:
            raise TypeError('{0} does not exist'.format(value))

    @property
    def target(self):
        if 'target' not in self.state:
            self.target = 'auto'

        return self.state['target']

    @target.setter
    def target(self, value):
        if value == 'auto':
            self.state['target'] = os.path.join(self.conf.paths.projectroot, self.conf.paths.source,
                                                self.state['_raw_source'])
            return
        elif value.startswith('/'):
            value = value[1:]

        if '{root}' in value:
            value = value.format(root=self.conf.paths.projectroot)

        if '{branch}' in value:
            value = value.format(branch=self.conf.git.brancehs.current)

        self.state['target'] = os.path.join(self.conf.paths.projectroot,
                                            self.conf.paths.source, value)

        self.state['_raw_target'] = value

    @property
    def transform(self):
        if 'transform' in self.state:
            return self.state['transform']
        else:
            return None

    @transform.setter
    def transform(self, value):
        if 'transform' in self.state:
            if isinstance(value, list):
                self.state['transform'].extend([TransformSpecification(v)
                                                for v in value])
            else:
                self.state['transform'].append(TransformSpecification(value))
        else:
            if isinstance(value, list):
                self.state['transform'] = [TransformSpecification(v)
                                           for v in value]
            else:
                self.state['transform'] = [TransformSpecification(value)]

    @property
    def truncate(self):
        if 'truncate' in self.state:
            return self.state['truncate']
        else:
            return None

    @truncate.setter
    def truncate(self, value):
        self.state['truncate'] = TruncateSpecification(value)

    @property
    def append(self):
        if 'append' in self.state:
            return self.state['append']
        else:
            return None

    @append.setter
    def append(self, value):
        if isinstance(value, basestring):
            self.state['append'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), str(value)))


class TransformSpecification(libgiza.config.ConfigurationBase):
    @property
    def regex(self):
        return self.state['regex']

    @regex.setter
    def regex(self, value):
        if isinstance(value, basestring):
            try:
                self.state['regex'] = re.compile(value)
            except:
                pass

        if 'regex' not in self.state:
            raise TypeError('{0}: {1}'.format(type(value), str(value)))

    @property
    def replace(self):
        return self.state['replace']

    @replace.setter
    def replace(self, value):
        if isinstance(value, basestring):
            self.state['replace'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), str(value)))


class TruncateSpecification(libgiza.config.ConfigurationBase):
    def ingest(self, input_obj):
        if input_obj is None:
            return

        if 'start-after' in input_obj:
            input_obj['start_after'] = input_obj['start-after']
            del input_obj['start-after']

        if 'end-before' in input_obj:
            input_obj['end_before'] = input_obj['end-before']
            del input_obj['end-before']

        super(TruncateSpecification, self).ingest(input_obj)

    @property
    def start_after(self):
        if 'start_after' in self.state:
            return self.state['start_after']
        else:
            return None

    @start_after.setter
    def start_after(self, value):
        if isinstance(value, basestring) or isinstance(value, int):
            self.state['start_after'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), str(value)))

    @property
    def end_before(self):
        if 'end_before' in self.state:
            return self.state['end_before']
        else:
            return None

    @end_before.setter
    def end_before(self, value):
        if isinstance(value, basestring) or isinstance(value, int):
            self.state['end_before'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), str(value)))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Development server in giza for more realistic local previews of rendered pages.
"""

import sys
import os.path
import logging
import argh

from giza.config.helper import fetch_config

logger = logging.getLogger('giza.operations.http')

if sys.version_info[0] == 2:
    import SocketServer as socket_server
    import SimpleHTTPServer as http_server
else:
    import socketserver as socket_server
    import http.server as http_server


class RequestHandler(http_server.SimpleHTTPRequestHandler):

    """Request handler wrapper that hosts files rooted at a particular
       directory."""

    def translate_path(self, path):
        """Map a request into the build/ directory."""
        path = os.path.relpath(path, '/')
        return os.path.join(self.root, path)

    def log_message(self, fmt, *args):
        """Pass this server event into the operation logger."""
        logger.info(fmt % args)


@argh.arg('--port', '-p', default=8090, dest='port')
@argh.arg('--builder', '-b', nargs='*', default='publish')
@argh.arg('--edition', '-e')
@argh.named('http')
@argh.expects_obj
def start(args):
    """Start an HTTP server rooted in the build directory."""
    conf = fetch_config(args)

    if conf.runstate.is_publish_target():
        RequestHandler.root = conf.paths.public_site_output
    elif conf.runstate.edition is not None:
        RequestHandler.root = os.path.join(conf.paths.projectroot,
                                           conf.paths.branch_output,
                                           '-'.join((args.builder[0], args.edition)))
    else:
        RequestHandler.root = os.path.join(conf.paths.projectroot,
                                           conf.paths.branch_output,
                                           args.builder[0])

    httpd = socket_server.TCPServer(('', conf.runstate.port), RequestHandler)
    logger.info('Hosting {0} at http://localhost:{1}/'.format(RequestHandler.root,
                                                              conf.runstate.port))
    httpd.serve_forever()



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Takes the ingested and processed data from all examples and renders the content
into reStructuredText using rstcloth.
"""

import logging

from rstcloth.rstcloth import RstCloth

logger = logging.getLogger('giza.content.examples.views')


def full_example(collection, examples):
    """
    :pram collection: An object with collection information and description.

    :pram examples:  An object with examples, procedures and results.

    See :mod:`giza.content.examples.modules` for full documentation of the
    example data format.

    :returns: A populated ``RstCloth()`` object with the content of one example.
    """

    r = RstCloth()

    if collection is not None:
        collection.render()

        if collection.options.show_title is True:
            if len(examples) == 1:
                ex_str = 'Example'
            else:
                ex_str = 'Examples'

            r.h2(ex_str)
            r.newline()

        if 'pre' in collection:
            r.content(collection.pre)
            r.newline()

        if 'content' in collection and collection.options.show_collection is True:
            r.content(collection.content)
            r.newline()

        if 'documents' in collection:
            r.codeblock(content=collection.documents,
                        language='javascript')
            r.newline()

        if 'post' in collection:
            r.content(collection.post)
            r.newline()

        if 'final' in collection:
            r.content(collection.final)
            r.newline()

    for idx, example in enumerate(examples):
        example.render()
        if idx != 0:
            r.newline(2)

        if len(examples) > 1 and 'title' in example:
            getattr(r, 'h' + str(example.title.level))(example.title.text)
            r.newline()

        if 'pre' in example:
            r.content(example.pre)
            r.newline()

        lang = set()
        for op in example.operation:
            if 'pre' in op:
                r.content(op.pre)
                r.newline()

            # if content in op (e.g. for literalincludes) then
            # no need for code or language
            if 'content' in op:
                r.content(op.content)
                r.newline()
            elif 'literalinclude' in op:  # Temporary and klugey
                include_options = []
                if 'language' in op:
                    include_options.append(('language', op.language))
                r.directive('literalinclude', op.literalinclude, include_options)
                r.newline()
            elif 'code' in op:
                lang.add(op.language)
                r.codeblock(content=op.code,
                            language=op.language)
                r.newline()

            if 'post' in op:
                r.content(op.post)
                r.newline()

        if 'post' in example:
            r.content(example.post)
            r.newline()

        if 'results' in example and example.results is not None:
            num_langs = len(lang)
            lang = list(lang)[0]
            if num_langs > 1:
                msg = 'specified more than one language for examples %s, using %s for results'
                logger.warning(msg, example.ref, lang)

            r.codeblock(content=example.results,
                        language=lang)

        if 'final' in example:
            r.newline()
            r.content(example.final)

    return r



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path

from giza.inheritance import DataContentBase, DataCache
from giza.content.glossary.models import GlossaryData


class GlossaryFile(DataContentBase):
    content_class = GlossaryData

    def target(self, fn):
        # fn is the source file
        return os.path.join(self.conf.system.content.glossary.output_dir,
                            self.conf.system.content.glossary.get_basename(fn)) + '.rst'


class GlossaryDataCache(DataCache):
    content_class = GlossaryFile
    content_type = 'glossary'



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import logging
import subprocess

import argh

from libgiza.git import GitRepo
from giza.config.code_review import CodeReviewConfiguration
from giza.config.helper import new_credentials_config

logger = logging.getLogger('giza.operations.code_review')


def safe_create_code_review_data_file(fn):
    if os.path.isfile(fn):
        return

    with open(fn, 'w') as f:
        json.dump({'branches': {}}, f)
        logger.info('created new code review cache')


def get_cr_data_file(arg):
    if isinstance(arg, GitRepo):
        path = arg.top_level()
    else:
        path = arg.paths.projectroot

    return os.path.join(path, '.git', 'code_review_mapping.json')


@argh.expects_obj
@argh.named('list')
def list_reviews(args):
    "Lists tracked code reviews."

    g = GitRepo()
    cr_data_file = get_cr_data_file(g)
    safe_create_code_review_data_file(cr_data_file)

    crconf = CodeReviewConfiguration(cr_data_file)

    print(json.dumps([k for k in crconf.branches.keys()], indent=3, sort_keys=True))


@argh.expects_obj
@argh.arg('_branch_name', nargs='*')
def close(args):
    "Removes a tracked code review."

    g = GitRepo()
    cr_data_file = get_cr_data_file(g)

    safe_create_code_review_data_file(cr_data_file)

    with CodeReviewConfiguration.persisting(cr_data_file) as data:
        branches = data.branches.keys()

        for to_delete in args._branch_name:
            if to_delete in branches:
                del data.branches[to_delete]
                logger.info('removed tracked code review for: ' + to_delete)
            else:
                logger.info("not tracking a code review for: " + to_delete)

            try:
                g.remove_branch(to_delete, args.force)
                logger.info('removed branch: ' + to_delete)
            except:
                logger.error('could not remove branch: ' + to_delete)


@argh.expects_obj
@argh.arg('_branch_name')
def checkout(args):
    "Checks out a tracked code review branch."

    g = GitRepo()
    cr_data_file = get_cr_data_file(g)

    safe_create_code_review_data_file(cr_data_file)
    crconf = CodeReviewConfiguration(cr_data_file)

    if args._branch_name in crconf.branches:
        try:
            g.checkout(args._branch_name)
            logger.info('checked out: ' + args._branch_name)
        except:
            logger.error('could not checkout branch: ' + args._branch_name)
    else:
        m = "no branch named {0} tracked. Please use another method to checkout this branch"
        logger.warning(m.format(args._branch_name))


@argh.named('send')
@argh.expects_obj
def create_or_update(args):
    "Creates or updates a code review case."

    creds = new_credentials_config()
    g = GitRepo()
    cr_data_file = get_cr_data_file(g)

    safe_create_code_review_data_file(cr_data_file)

    with CodeReviewConfiguration.persisting(cr_data_file) as data:
        if g.current_branch() in data.branches:
            cr_data = data.branches[g.current_branch()]
            if len(cr_data.commits) > 1 and g.sha('HEAD~') == cr_data.commits[-2]:
                # the last commit was amended, so replace it:
                if g.sha('HEAD') != cr_data.commits[-1]:
                    cr_data.commits[-1] = g.sha('HEAD')
            elif g.sha() not in cr_data.commits:
                cr_data.commits.append(g.sha())

            if len(cr_data.commits) >= 2:
                use_hash = str('..'.join([cr_data.commits[0][0:8], cr_data.commits[-1][0:8]]))
            else:
                use_hash = str(cr_data.commits[-1])

            logger.info('updating an existing code review.')
            update_code_review(cr_data, g, use_hash)
        else:
            data.set_branch(g.current_branch(), {'original_name': g.commit_messages()[0],
                                                 'commits': [g.sha('HEAD~'), g.sha()]})

            logger.info('creating new code review.')
            create_code_review(data, g, creds)

# Worker functions to create or update code reviews


def update_code_review(cr_data, g, use_hash):
    cmd = [
        'upload.py',
        '--oauth2',
        '-y',
        '--nojira',
        '--email', g.author_email(),
        '-m', '"' + cr_data.original_name + '"',
        '-i', cr_data.issue
    ]

    if len(cr_data.commits) > 2:
        cmd.append('--rev')

    cmd.append(use_hash)

    logger.debug(' '.join(cmd))

    try:
        cr_upload = subprocess.check_output(cmd, stderr=subprocess.STDOUT).strip()
        issue_url = get_issue_url(cr_upload)
        logger.info('updated issue: ' + issue_url)
    except subprocess.CalledProcessError as e:
        print(e.returncode)
        print(e.output)
        print(' '.join(cmd))
        logger.error('failed to update issue')


def create_code_review(data, g, creds):
    branch_data = data.get_branch(g.current_branch())

    cmd = ['upload.py',
           '--oauth2',
           '-y']

    if creds is not None:
        cmd.extend(['--jira_user', creds.jira.username])

    cmd.extend(['--email', g.author_email(),
                '-m', "'" + g.commit_messages()[0] + "'",
                '..'.join(branch_data.commits)])

    try:
        logger.info(' '.join(cmd))
        cr_upload = subprocess.check_output(cmd, stderr=subprocess.STDOUT).strip()

        branch_data.issue = get_issue_number(cr_upload)
        issue_url = get_issue_url(cr_upload)

        data.set_branch(g.current_branch(), branch_data)

        if len(issue_url) < 80:
            logger.info('created issue: ' + issue_url)
        else:
            logger.info('created new code review issue')
    except subprocess.CalledProcessError as e:
        del data.branches[g.current_branch()]
        logger.error('failed to create issue')
        logger.error(' '.join(cmd))
        print(e.returncode, e.output)

# Output processing


def get_issue_url(output):
    if not isinstance(output, list):
        output = output.split('\n')

    urls = []
    for ln in output:
        if 'http' in ln:
            urls.append(ln.split(' ')[-1])

    if len(urls) > 1:
        urls = urls[1:]

    return '\n'.join(urls)


def get_issue_number(output):
    if not isinstance(output, list):
        output = output.split('\n')

    for ln in output:
        if ln.startswith("Issue "):
            return ln.rsplit('/', 1)[-1]

    # if we get here, we're in trouble:
    logger.error('did not create an issue')
    print('\n'.join(output))
    raise Exception



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import itertools
import logging

import libgiza.config
import giza.tools.files

logger = logging.getLogger('giza.config.')


class ContentType(libgiza.config.ConfigurationBase):
    _option_registry = ['dir']

    @property
    def sources(self):
        files = giza.tools.files.expand_tree(self.dir, 'yaml')

        sources = set()
        for prefix, fn in itertools.product(self.prefixes, files):
            prefix = os.path.join(self.dir, prefix)
            if fn.startswith(prefix):
                sources.add(fn)

        return list(sources)

    @property
    def prefixes(self):
        if 'prefixes' not in self.state:
            self.prefixes = []

        return self.state['prefixes']

    @prefixes.setter
    def prefixes(self, value):
        if 'prefixes' not in self.state:
            self.state['prefixes'] = []

        if not value or value is None:
            return

        if not isinstance(value, list):
            value = [value]

        for prefix in value:
            if prefix in self.state['prefixes']:
                continue
            else:
                self.state['prefixes'].append(prefix)

    @property
    def name(self):
        return self.state['name']

    @name.setter
    def name(self, value):
        self.state['name'] = value
        self.prefixes = value

    @property
    def fn_prefix(self):
        return os.path.join(self.dir, self.name)

    def get_basename(self, fn):
        return fn[len(self.fn_prefix) + 1:-5]

    @property
    def output_dir(self):
        if 'output_dir' in self.state:
            dirname = self.state['output_dir']
        else:
            dirname = self.name

        return os.path.join(self.dir, dirname)

    @output_dir.setter
    def output_dir(self, value):
        self.state['output_dir'] = value

    @property
    def task_generator(self):
        if '_task_generator' in self.state:
            return self.state['_task_generator']
        else:
            logger.warning('returning no-op content generator for: ' + self.name)

            def nop(*args, **kwargs):
                pass
            return nop

    @task_generator.setter
    def task_generator(self, value):
        if callable(value):
            self.state['_task_generator'] = value
        else:
            raise TypeError


class ContentRegistry(libgiza.config.ConfigurationBase):

    def add(self, name, definition):
        if not isinstance(definition, ContentType):
            raise TypeError
        else:
            if name not in self.state:
                self._option_registry.append(name)
                self.state[name] = definition

    def get(self, name):
        if name in self.state:
            return self.state[name]
        else:
            raise AttributeError(name)

    def iterator(self):
        return self.state.values()

    def output_directories(self, prefix_len=0):
        for content_type in self.iterator():
            yield content_type.output_dir[prefix_len:]

    @property
    def task_generators(self):
        for content in self.state.values():
            yield content, content.task_generator

    @property
    def content_prefixes(self):
        for name, content in self.state.items():
            yield name, content.prefixes

# Factories


def new_content_type(name, conf, task_generator=None, source_dir=None,
                     output_dir=None, prefixes=None):
    if source_dir is None:
        source_dir = os.path.join(conf.paths.projectroot,
                                  conf.paths.branch_includes)

    if output_dir is None:
        output_dir = os.path.join(source_dir, name)

    c = ContentType()
    c.name = name
    c.dir = source_dir
    c.output_dir = output_dir

    if prefixes is not None:
        c.prefixes = prefixes

    if task_generator is not None:
        c.task_generator = task_generator

    return c



from nose.tools import nottest, istest

import os

# this runs tests of the inheritance.py baseclasses, as is.
from libgiza.test.test_inheritance import (TestDataCache, TestDataContentBase,
                                           TestInheritedContentResolution,
                                           TestBaseTemplateRendering,
                                           get_test_file_path,
                                           get_inheritance_data_files)

from giza.config.main import Configuration
from giza.config.runtime import RuntimeStateConfig

import libgiza.test
import libgiza.git

import giza.inheritance
import giza.config.git

import giza.content.steps.inheritance
import giza.content.steps.models
import giza.content.options.inheritance
import giza.content.options.models
import giza.content.tocs.inheritance
import giza.content.tocs.models
import giza.content.extract.inheritance
import giza.content.extract.models
import giza.content.release.inheritance
import giza.content.release.models
import giza.content.examples.inheritance
import giza.content.examples.models

def get_local_data_path():
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')

class TestGizaDataCache(TestDataCache):
    def setUp(self):
        self.c = Configuration()
        self.c.project = {'name': 'test'}
        self.c.runstate = RuntimeStateConfig()
        self.c.state['git'] = giza.config.git.GitConfig({}, self.c, os.getcwd())

        self.setUpClasses()
        self.create_data()

    def setUpClasses(self):
        path = get_test_file_path()
        self.c.paths = {'includes': path,
                        'source': path,
                        'projectroot': path,
                        'output': path}

        self.DataContentBase = giza.inheritance.DataContentBase
        self.DataCache = giza.inheritance.DataCache

class TestGizaDataContentBase(TestDataContentBase):
    def setUp(self):
        self.c = Configuration()
        self.c.project = {'name': 'test'}
        self.c.runstate = RuntimeStateConfig()
        self.c.state['git'] = giza.config.git.GitConfig({}, self.c, os.getcwd())

        self.setUpClasses()
        self.create_data()

    def setUpClasses(self):
        path = get_test_file_path()
        self.c.paths = {'includes': path,
                        'projectroot': path}
        self.dummy_doc = {'ref': 'dummy-doc',
                          'pre': 'pre text'}

        self.content_fn = get_inheritance_data_files()[0]
        self.DataContentBase = giza.inheritance.DataContentBase
        self.DataCache = giza.inheritance.DataCache
        self.InheritableContentBase = giza.inheritance.InheritableContentBase

class TestGizaInheritedContentResolution(TestInheritedContentResolution):
    def setUp(self):
        self.c = Configuration()
        self.c.project = {'name': 'test'}
        self.c.runstate = RuntimeStateConfig()
        self.c.state['git'] = giza.config.git.GitConfig({}, self.c, os.getcwd())

        self.setUpClasses()
        self.create_data()

    def setUpClasses(self):
        path = get_test_file_path()
        self.c.paths = {'includes': path,
                        'source': path,
                        'output': path,
                        'projectroot': path}

        self.DataCache = giza.inheritance.DataCache

class TestGizaBaseTemplateRendering(TestBaseTemplateRendering):
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.c.state['git'] = giza.config.git.GitConfig({}, self.c, os.getcwd())
        self.setUpClasses()
        self.create_data()

    def setUpClasses(self):
        self.InheritableContentBase = giza.inheritance.InheritableContentBase

# Base Classes for Reuse

@nottest
class GizaDataCacheBase(TestGizaDataCache):
    def create_data(self):
        path = get_local_data_path()
        self.c.paths = {'includes': path,
                        'source': path,
                        'projectroot': path,
                        'output': path}

        self.files = [os.path.join(get_local_data_path(), fn)
                      for fn in (self.short_name + '-one.yaml',
                                 self.short_name + '-two.yaml')]

        self.data = self.DataCache([], self.c)

@nottest
class GizaDataContentBase(TestGizaDataContentBase):
    def create_data(self):
        self.content_fn = os.path.join(get_local_data_path(),
                                       self.short_name + '-one.yaml')

        self.data = self.DataCache([self.content_fn], self.c)
        self.content = self.data.cache[self.content_fn]

@nottest
class GizaInheritedContentResolutionBase(TestGizaInheritedContentResolution):
    def create_data(self):
        path = get_local_data_path()
        self.c.paths = {'includes': path,
                        'source': path,
                        'projectroot': path,
                        'output': path}

        self.files = [os.path.join(self.c.paths.source, fn)
                      for fn in (self.short_name + '-one.yaml',
                                 self.short_name + '-two.yaml')]

        self.data = self.DataCache(self.files, self.c)


# Pass-through Tests of Steps

@istest
class TestStepsDataCache(GizaDataCacheBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.steps.inheritance.StepFile
        self.DataCache = giza.content.steps.inheritance.StepDataCache
        self.short_name = 'steps'


@istest
class TestStepsDataContentBase(GizaDataContentBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.steps.inheritance.StepFile
        self.DataCache = giza.content.steps.inheritance.StepDataCache
        self.InheritableContentBase = giza.content.steps.models.StepData
        self.short_name = 'steps'

        self.dummy_doc = {'ref': 'dummy-doc',
                          'pre': 'pre text'}


@istest
class TestStepsInheritedContentResolution():
    def setUpClasses(self):
        self.DataCache = giza.content.steps.inheritance.StepDataCache
        self.short_name = 'steps'
        self.len_source_docs = 9
        self.num_docs = 2

# Pass-through Tests of Options

@istest
class TestOptionsDataCache(GizaDataCacheBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.options.inheritance.OptionFile
        self.DataCache = giza.content.options.inheritance.OptionDataCache
        self.short_name = 'options'


@istest
class TestOptionsDataContentBase(GizaDataContentBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.options.inheritance.OptionFile
        self.DataCache = giza.content.options.inheritance.OptionDataCache
        self.InheritableContentBase = giza.content.options.models.OptionData
        self.short_name = 'options'

        self.dummy_doc = {'name': 'dummydoc',
                          'program': 'test',
                          'pre': 'pre text'}

@istest
class TestOptionsInheritedContentResolution(GizaInheritedContentResolutionBase):
    def setUpClasses(self):
        self.DataCache = giza.content.options.inheritance.OptionDataCache
        self.short_name = 'options'
        self.len_source_docs = 9
        self.num_docs = 2

# Pass-through Tests of Tocs

@istest
class TestTocsDataCache(GizaDataCacheBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.tocs.inheritance.TocFile
        self.DataCache = giza.content.tocs.inheritance.TocDataCache
        self.short_name = 'toc'

@istest
class TestTocsDataContentBase(GizaDataContentBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.tocs.inheritance.TocFile
        self.DataCache = giza.content.tocs.inheritance.TocDataCache
        self.InheritableContentBase = giza.content.tocs.models.TocData
        self.short_name = 'toc'

        self.dummy_doc = {'name': 'dummydoc',
                          'file': '/path/to/doc',
                          'description': 'text'}

@istest
class TestTocsInheritedContentResolution(GizaInheritedContentResolutionBase):
    def setUpClasses(self):
        self.DataCache = giza.content.tocs.inheritance.TocDataCache
        self.short_name = 'toc'
        self.len_source_docs = 7
        self.num_docs = 2


# Pass-through Tests of Extracts

@istest
class TestExtractsDataCache(GizaDataCacheBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.extract.inheritance.ExtractFile
        self.DataCache = giza.content.extract.inheritance.ExtractDataCache
        self.short_name = 'extracts'



@istest
class TestExtractsDataContentBase(GizaDataContentBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.extract.inheritance.ExtractFile
        self.DataCache = giza.content.extract.inheritance.ExtractDataCache
        self.InheritableContentBase = giza.content.extract.models.ExtractData
        self.short_name = 'extracts'

        self.dummy_doc = {'ref': 'dummy-doc',
                          'pre': 'pre text'}


@istest
class TestExtractsInheritedContentResolution(GizaInheritedContentResolutionBase):
    def setUpClasses(self):
        self.DataCache = giza.content.extract.inheritance.ExtractDataCache
        self.short_name = 'extracts'
        self.len_source_docs = 4
        self.num_docs = 2


# Pass-through Tests of Releases

@istest
class TestReleasesDataCache(GizaDataCacheBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.release.inheritance.ReleaseFile
        self.DataCache = giza.content.release.inheritance.ReleaseDataCache
        self.short_name = 'release'


@istest
class TestReleasesDataContentBase(GizaDataContentBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.release.inheritance.ReleaseFile
        self.DataCache = giza.content.release.inheritance.ReleaseDataCache
        self.InheritableContentBase = giza.content.release.models.ReleaseData
        self.short_name = 'release'

        self.dummy_doc = {'ref': 'dummy-doc',
                          'pre': 'pre text'}


@istest
class TestReleasesInheritedContentResolution(GizaInheritedContentResolutionBase):
    def setUpClasses(self):
        self.DataCache = giza.content.release.inheritance.ReleaseDataCache
        self.short_name = 'release'
        self.len_source_docs = 5
        self.num_docs = 2


# Pass-through Tests of Examples

@istest
class TestExamplesDataCache(GizaDataCacheBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.examples.inheritance.ExampleFile
        self.DataCache = giza.content.examples.inheritance.ExampleDataCache
        self.short_name = 'examples'


@istest
class TestExamplesDataContentBase(GizaDataContentBase):
    def setUpClasses(self):
        self.DataContentBase = giza.content.examples.inheritance.ExampleFile
        self.DataCache = giza.content.examples.inheritance.ExampleDataCache
        self.InheritableContentBase = giza.content.examples.models.ExampleData
        self.short_name = 'examples'

        self.dummy_doc = {'ref': 'dummy-doc',
                          'pre': 'pre text',
                          'operation': { 'language': 'javascript',
                                         'code': ['use db']}}


@istest
class TestExamplesInheritedContentResolution(GizaInheritedContentResolutionBase):
    def setUpClasses(self):
        self.DataCache = giza.content.examples.inheritance.ExampleDataCache
        self.short_name = 'examples'
        self.len_source_docs = 9
        self.num_docs = 2



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import logging
import itertools
import copy

import yaml

from giza.config.main import Configuration
from giza.config.error import ConfigurationError
from giza.config.runtime import RuntimeStateConfig
from giza.config.project import get_path_prefix
from giza.config.credentials import CredentialsConfig, get_credentials_skeleton
from giza.config.sphinx_config import render_sconf

from giza.content.release.tasks import register_releases
from giza.content.extract.tasks import register_extracts
from giza.content.options.tasks import register_options
from giza.content.examples.tasks import register_examples
from giza.content.steps.tasks import register_steps
from giza.content.glossary.tasks import register_glossary
from giza.content.tocs.tasks import register_toc
from giza.content.apiargs.tasks import register_apiargs
from giza.content.changelog.tasks import register_changelogs

logger = logging.getLogger('giza.config.helper')


def new_credentials_config(conf_path=None):
    if conf_path is None:
        for fn in [os.path.expanduser("~/.giza-credentials.yaml"),
                   os.path.expanduser("~/.mongodb-jira.yaml")]:
            if os.path.isfile(fn):
                conf_path = fn
                break

    if conf_path is None and not (os.path.isfile(conf_path) or isinstance(conf_path, dict)):
        return None
    else:
        return CredentialsConfig(conf_path)


def fetch_config(args):
    c = Configuration()
    c.ingest(args.conf_path)
    c.runstate = args

    register_content_generators(c)

    return c


def register_content_generators(conf):
    logger.debug("registering content generators with config")
    register_options(conf)
    register_steps(conf)
    register_glossary(conf)
    register_releases(conf)
    register_examples(conf)
    register_extracts(conf)
    register_toc(conf)
    register_apiargs(conf)
    register_changelogs(conf)


def new_skeleton_config(conf=None):
    if conf is None:
        conf = Configuration()
        args = RuntimeStateConfig()
        conf.runstate = args

        try:
            register_content_generators(conf)
        except KeyError:
            logger.warning("trouble registering content generators")

        return conf
    elif isinstance(conf, RuntimeStateConfig):
        return fetch_config(conf)
    else:
        return conf


def setup_credentials(args):
    skel = get_credentials_skeleton()

    dump_skel(skel, args)


def new_config(args=None):
    if args in (None, True, False):
        args = RuntimeStateConfig()

        return fetch_config(args)
    elif isinstance(args, RuntimeStateConfig):
        return fetch_config(args)
    elif isinstance(args, Configuration):
        return args
    else:
        raise ConfigurationError


def dump_skel(skel, args):
    conf_path = os.path.expanduser(os.path.join("~", args.user_conf_path))
    if os.path.exists(conf_path) and args.force is False:
        logger.error('{0} already exists. exiting.'.format(conf_path))
        exit(1)

    with open(conf_path, 'w') as f:
        yaml.dump(skel, f, default_flow_style=False)
        f.write('...\n')
        logger.info('wrote scrumpy configuration skeleton to: {0}')


def get_builder_jobs(conf):
    # assemble a list to generate tasks in the form of:
    # ((edition, language, builder), (conf, sconf))

    for edition, language, builder in itertools.product(conf.runstate.editions_to_build,
                                                        conf.runstate.languages_to_build,
                                                        conf.runstate.builder):
        yield ((edition, language, builder),
               get_sphinx_build_configuration(edition, language, builder, conf.runstate))


def get_restricted_builder_jobs(conf):
    build_source_copies = set()

    for ((edition, langauge, builder), (bconf, sconf)) in get_builder_jobs(conf):
        if bconf.paths.branch_source not in build_source_copies:
            build_source_copies.add(bconf.paths.branch_source)

            yield ((edition, langauge, builder), (bconf, sconf))


def get_sphinx_build_configuration(edition, language, builder, args):
    """
    Given an ``edition``, ``language`` and ``builder`` strings and the runtime
    arguments, return copies of the configuration (``conf``) and sphinx
    configuration (``sconf``) objects.
    """
    args._conf = None
    args = copy.deepcopy(args)

    args.language = language
    args.edition = edition
    args.builder = builder

    conf = fetch_config(args)
    args.conf = conf
    sconf = render_sconf(edition, builder, language, conf)

    return conf, sconf


def get_manual_path(conf):
    if conf.system.branched is False:
        return conf.project.tag
    else:
        branch = conf.git.branches.current
        return get_path_prefix(conf, branch)


def get_versions(conf):
    o = []

    current_branch = conf.git.branches.current

    if current_branch not in conf.git.branches.published:
        current_version_index = 0
    else:
        current_version_index = conf.git.branches.published.index(current_branch)

    for idx, version in enumerate(conf.version.published):
        v = {}

        branch = conf.git.branches.published[idx]
        v['path'] = get_path_prefix(conf, branch)

        v['text'] = version
        if version == conf.version.stable:
            v['text'] += ' (current)'

        if version == conf.version.upcoming:
            v['text'] += ' (upcoming)'

        v['current'] = True if idx == current_version_index else False

        o.append(v)

    return o



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import collections

import giza.jeerah.client
import rstcloth.rstcloth as rstcloth

logger = logging.getLogger('giza.content.changelog.views')


def migrate_changelog(fn, conf):
    # because changelogs are generated in the source working dir we need to
    # migrate it into the build directory, and will use this as a finalizer.

    # filename relative to the top of the source directory
    base = fn[len(os.path.join(conf.paths.projectroot, conf.paths.source))+1:]
    target = os.path.join(conf.paths.projectroot, conf.paths.branch_source, base)

    giza.tools.files.copy_if_needed(fn, target, "changelog")


def get_issue_structure(version, conf):
    """
    Collect data from jira and produce a special headings structure for this version.
    """

    # set up and connect to jira
    jira = giza.jeerah.client.JeerahClient(conf)
    jira.connect()

    # run the jira query
    projects = ', '.join(conf.system.files.data.jira.site.projects)
    query = "project in ({0}) and fixVersion = {1} and resolution = 'Fixed' ORDER BY key ASC".format(projects, version)
    issues = jira.query(query)
    logger.info("building changelog for {0} with {1} issue(s)".format(version, len(issues)))

    # setup container of heading groups using the defined ordering.
    headings = collections.OrderedDict()
    for k in conf.system.files.data.jira.changelog.ordering:
        headings[k] = list()

    # invert the mapping of groups to components so we can filter
    groups = dict()
    for k, v in conf.system.files.data.jira.changelog.groups.items():
        for c in v:
            groups[c] = k

    # run through all issues, and put each one in the headings structure at the
    # best place
    for issue in issues:
        components = []
        for c in issue.fields.components:
            components.append(c.name)

            if c.name not in groups:
                logger.error("undefined component %s. update configuration before continuing", c.name)

        issue_pair = (issue.key.encode("utf-8"), issue.fields.summary.encode("utf-8"))

        if len(components) == 0:
            # if there isn't a component put this in the last grouping.
            headings[next(reversed(headings))].append(issue_pair)
        elif len(components) == 1:
            headings[groups[components[0]]].append(issue_pair)
        else:
            # if an issue has multiple components use the one that appears first
            # in the ordering of headings
            located = False
            for heading in groups:
                if heading in components:
                    headings[groups[heading]].append(issue_pair)
                    located = True
                    break

            # if we get here, we should stop the build because someone added a
            # ticket with a component and we don't know how to deal with this.
            if located is False:
                logger.error("skipping issue {0} in changelog {1} because its components aren't defined in the changelog configuration. Fix now.".format(issue_pair[0], version))
                raise SystemExit

    return headings


def get_changelog_content(fn, version, conf):
    """
    Builds and writes the heading based on the ordered mapping of heading groups
    to RST files in the source directory.
    """

    # this queries jira and builds an map[OrderedDict]list<issue_pairs>
    # structure that holds the data and groupings
    headings = get_issue_structure(version, conf)

    # invert the mapping of nested, so we can properly handle subheadings.
    nested = dict()
    for enclosing_level, sub_headings in conf.system.files.data.jira.changelog.nesting.items():
        for component in sub_headings:
            nested[component] = enclosing_level

    # build the changelog content itself.
    r = rstcloth.RstCloth()
    level = 3

    # headings and links
    r.ref_target("{0}-changelog".format(version))
    r.newline()
    r.heading(text="{0} Changelog".format(version), char=giza.content.helper.character_levels[level-1])
    r.newline()

    # process all of the issues by group.
    for heading, issues in headings.items():
        if heading in nested:
            # we deal with nested headings when we do their parent. skip here.
            continue
        else:
            if heading in conf.system.files.data.jira.changelog.nesting and len(issues) == 0:
                # if a heading has subheadings, and all are empty, then we should skip it entirely.
                empty_sub_headings = 0
                for sub in conf.system.files.data.jira.changelog.nesting[heading]:
                    if len(headings[sub]) == 0:
                        empty_sub_headings += 1
                if empty_sub_headings == len(conf.system.files.data.jira.changelog.nesting[heading]):
                    continue
            elif len(issues) == 0:
                # skip empty headings.
                continue

            # format the heading.
            r.heading(text=heading, indent=0,
                      char=giza.content.helper.character_levels[level])
            r.newline()

            if len(issues) == 1:
                r.content("{1} {0}".format(issues[0][1], r.role("issue", issues[0][0])), wrap=False)
            else:
                for issue in issues:
                    r.li("{1} {0}".format(issue[1], r.role("issue", issue[0])), wrap=False)
            r.newline()

            # repeat the above formatting with minor variations to do the nesting.
            if heading in conf.system.files.data.jira.changelog.nesting:
                for sub in conf.system.files.data.jira.changelog.nesting[heading]:
                    if len(headings[sub]) == 0:
                        continue

                    r.heading(text=sub, indent=0,
                              char=giza.content.helper.character_levels[level+1])
                    r.newline()

                    sub_issues = headings[sub]
                    if len(sub_issues) == 0:
                        r.content("{1} {0}".format(sub_issues[0][1].strip(), r.role("issue", sub_issues[0][0])), wrap=False)
                    else:
                        for issue in sub_issues:
                            r.li("{1} {0}".format(issue[1].strip(), r.role("issue", issue[0])), wrap=False)
                    r.newline()

    r.write(fn)
    logger.info("wrote changelog '{0}'. Commit this file independently.".format(fn))
    migrate_changelog(fn, conf)


def render_intermediate_files(fn, version, releases, conf):
    """
    Each major release series has an "intermediate" file that includes the
    changelog file for each release. This way, there's no requirement to change
    the changelog file every release.
    """

    r = rstcloth.RstCloth()
    for rel in releases:
        r.directive("include", "/includes/changelogs/releases/{0}.rst".format('.'.join([str(s) for s in rel])))
        r.newline()

    r.write(fn)
    migrate_changelog(fn, conf)
    logger.info("wrote intermediate versions file ")



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import libgiza.config

logger = logging.getLogger('giza.config.')


def get_path_prefix(conf, branch):
    """
    Returns the part of a site path between the domain name and the paths
    produced from sprint, accounting for sub-site names and branch names.
    """

    o = []

    if conf.project.siteroot is True:
        if (conf.project.branched is True and
                conf.git.branches.manual != branch):
            o.append(branch)
        else:
            o.append(conf.project.tag)
    else:
        if conf.project.basepath not in ('', None):
            o.append(conf.project.basepath)

        if conf.project.branched is True:
            if (branch == conf.git.branches.current and
                    conf.git.branches.manual == conf.git.branches.current):
                o.append('current')
            else:
                o.append(branch)

    return '/'.join(o)


def get_current_path(conf):
    branch = conf.git.branches.current
    if branch not in conf.git.branches.published:
        branch = conf.git.branches.published[0]

    return get_path_prefix(conf, branch)


class ProjectConfig(libgiza.config.RecursiveConfigurationBase):
    _option_registry = ['name', 'title']

    @property
    def tag(self):
        if self.edition in self.edition_list:
            return self.edition_map[self.edition].tag
        elif 'tag' in self.state:
            return self.state['tag']
        else:
            return ''

    @tag.setter
    def tag(self, value):
        self.state['tag'] = value

    @property
    def editions(self):
        if 'editions' in self.state:
            return self.state['editions']
        else:
            return []

    @property
    def edition_list(self):
        if '_edition_list' in self.state:
            return self.state['_edition_list']
        else:
            return []

    @property
    def edition_map(self):
        if '_edition_map' in self.state:
            return self.state['_edition_map']
        else:
            return {}

    @editions.setter
    def editions(self, value):
        if isinstance(value, list):
            if '_edition_list' not in self.state:
                self.state['_edition_list'] = []
            if 'editions' not in self.state:
                self.state['editions'] = []
            if '_edition_map' not in self.state:
                self.state['_edition_map'] = {}

            for v in value:
                ename = v['name']
                ed = EditionListConfig(v)
                self.state['_edition_list'].append(ename)
                self.state['editions'].append(ed)
                self.state['_edition_map'][ename] = ed
        else:
            logger.critical('editions must be a list')
            raise TypeError

    @property
    def edition(self):
        if 'edition' not in self.state or self.state['edition'] is None:
            self.edition = None
            if 'edition' not in self.state:
                return self.name
            else:
                return self.state['edition']
        else:
            return self.state['edition']

    @edition.setter
    def edition(self, value):
        if 'editions' in self.state:
            if value in self.state['_edition_list']:
                self.state['edition'] = value
            elif self.conf.runstate.edition in self.state['_edition_list']:
                self.state['edition'] = self.conf.runstate.edition

    @property
    def branched(self):
        if 'branched' not in self.state:
            self.branched = None

        return self.state['branched']

    @branched.setter
    def branched(self, value):
        if isinstance(value, bool):
            self.state['branched'] = value
        else:
            self.state['branched'] = False
            for edition in self.editions:
                if self.edition == edition.name:
                    self.state['branched'] = edition.branched
                    break

    @property
    def url(self):
        if 'url' not in self.state:
            self.url = None

        return self.state['url']

    @url.setter
    def url(self, value):
        url = None
        for edition in self.editions:
            if self.edition == edition.name:
                url = edition.url
                break

        if url is None:
            if value is None:
                self.state['url'] = ''
            else:
                self.state['url'] = value
        else:
            self.state['url'] = url

    @property
    def basepath(self):
        if 'basepath' not in self.state:
            self.basepath = None

        if 'basepath' in self.state:
            return self.state['basepath']
        else:
            return ''

    @basepath.setter
    def basepath(self, value):
        if value is not None:
            self.state['basepath'] = value
        else:
            for edition in self.editions:
                if self.edition == edition.name:
                    if edition.tag is None:
                        self.state['basepath'] = ''
                    else:
                        self.state['basepath'] = edition.tag

                    break

            if 'basepath' not in self.state and 'tag' in self.state:
                self.state['basepath'] = self.tag

    @property
    def siteroot(self):
        if 'siteroot' in self.state:
            return self.state['siteroot']
        else:
            return False

    @siteroot.setter
    def siteroot(self, value):
        if isinstance(value, bool):
            self.state['siteroot'] = value
        else:
            self.state['siteroot'] = bool(value)

    @property
    def sitepath(self):
        return get_path_prefix(self.conf, self.conf.git.branches.current)


class EditionListConfig(libgiza.config.ConfigurationBase):
    _option_registry = ['name', 'url']

    @property
    def branched(self):
        if 'branched' in self.state:
            return self.state['branched']
        else:
            return False

    @branched.setter
    def branched(self, value):
        if isinstance(value, bool):
            self.state['branched'] = value
        else:
            raise TypeError

    @property
    def tag(self):
        if 'tag' not in self.state:
            return None
        else:
            return self.state['tag']

    @tag.setter
    def tag(self, value):
        self.state['tag'] = value



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple operations to create and bootstrap the content for new projects.
"""

import inspect
import logging
import os
import shutil
import subprocess

import argh
import giza

from libgiza.app import BuildApp
from libgiza.git import GitRepo, GitError

from giza.config.helper import fetch_config
from giza.operations.sphinx_cmds import sphinx_publication
from giza.tools.files import safe_create_directory

logger = logging.getLogger('giza.operations.quickstart')


@argh.arg('--with-git', action='store_true', dest='quickstart_git')
@argh.named('quickstart')
@argh.expects_obj
def make_project(args):
    """
    Generate a project skeleton. Prefer this operation over
    ``sphinx-quickstart``. Also builds skeleton HTML artifacts.
    """
    if args.quickstart_git is True:
        logger.info('creating a new git repository')
        g = GitRepo(os.getcwd())
        g.create_repo()
        build_sphinx = True
    else:
        try:
            GitRepo().sha()
            build_sphinx = True
        except GitError:
            build_sphinx = False

    mod_path = os.path.dirname(inspect.getfile(giza))
    qstart_path = os.path.join(mod_path, 'quickstart')

    cmd = 'rsync --ignore-existing --recursive {0}/. {1}'.format(qstart_path, os.getcwd())
    r = subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)
    logger.info('migrated new site files')

    if args.quickstart_git is True:
        if not r.startswith('Reinitialized'):
            g.cmd('add', '-A')

            try:
                g.cmd('commit', '-m', '"initial commit"')
            except GitError:
                build_sphinx = False
                pass

    if build_sphinx is True:
        test_build_site(args)


def test_build_site(args):
    args.languages_to_build = args.editions_to_build = []
    args.builder = 'html'

    conf = fetch_config(args)

    safe_create_directory('build')
    with BuildApp.new(pool_type=conf.runstate.runner,
                      pool_size=conf.runstate.pool_size,
                      force=conf.runstate.force).context() as app:
        try:
            sphinx_publication(conf, args, app)
        except:
            sphinx_publication(conf, args, app)
            if os.path.exists('doc-tools'):
                shutil.rmtree('docs-tools')

    logger.info('bootstrapped makefile system')

    logger.info('updated project skeleton in current directory.')
